import numpy as np
from sklearn.linear_model import LinearRegression as LR
from sklearn.linear_model import LassoCV
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestRegressor as RF
from sklearn.ensemble import GradientBoostingRegressor as GB
from sklearn import metrics
from sklearn.cluster import SpectralClustering as SC
from genData import gen0, gen1, gen2, gen3
from sklearn import preprocessing
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from typing import List, Dict, Tuple



def getClust(W, k, PRINT=True):
	'''
		sklearn spectrum clustering, which does not seem to work well

		Input:
			W: distance matrix of modules
			k: number of clusters
		Output:
			lab: assigned cluster label for each module (by index)
	'''
	clu = SC(n_clusters=k, affinity='precomputed', assign_labels="discretize").fit(W)
	lab = clu.labels_
	if PRINT:
		print('lab', lab)

	return lab, k

def getClust_custom(W, k=None, PRINT=True):
	'''
		handwritten spectrum clustering, for understanding what is going on

		The adjacency matrix is used to compute a normalized graph Laplacian
	    whose spectrum (especially the eigenvectors associated to the
	    smallest eigenvalues) has an interpretation in terms of minimal
	    number of cuts necessary to split the graph into comparably sized
	    components.
	    This embedding can also 'work' even if the ``adjacency`` variable is
	    not strictly the adjacency matrix of a graph but more generally
	    an affinity or similarity matrix between samples (for instance the
	    heat kernel of a euclidean distance matrix or a k-NN matrix).
	    In our case, is defined by CV errors.

		Input:
			W: distance matrix of modules
			k: number of clusters
				if None: automatic selection by GAP statistics
		Output:
			lab: assigned cluster label for each module (by index)
			k: number of clusters (selected, if the input is None)
	'''

	n = W.shape[0]
	if k is not None and k > n:
		sys.exit('number of clusters too large!')
	if n < 3:
		sys.exit('number of vertices in W must be at least 3!')

	d = np.sum(W, axis=1)
	D = np.diag(d)
	D_temp = np.diag(1.0 / np.sqrt(d))
	L = D - W 
	L_sym = D_temp.dot(L).dot(D_temp)

	# eigval, eigvec are the eigvalues and normalized (unit “length”) eigenvectors
	# arranged in columns
	eigval, eigvec = np.linalg.eig(L_sym)

	# sort eigvals from small to large
	srt = np.argsort(np.abs(eigval))
	eigval = eigval[srt]
	eigvec = eigvec[:, srt]

	if k is None:
		# use gap statistics to select the k
		# the following ratio should expect a sudden increase in value
		# find the first i such that eigval[i+1]/eigval[i+2] is very large
		# is 'very' is defined by eigval[i+1]/eigval[i+2] > 0.8
		# the clustering should then correspond to 0,...,i, i.e. k=i+1
		k = n
		rat = eigval[:-1] / eigval[1:]
		for j in range(n-2):
			if rat[j+1] > 0.8:
				k = j + 1
				break
		if PRINT:
			print('n={}, and gap statistics selects k={}'.format(n, k))

	U = eigvec[:, :k]
	# normalize rows of U
	row_sums = np.sqrt(np.sum(np.square(U), axis=1))
	U = U / row_sums[:, np.newaxis]
	# U shape: (n_samples, n_features)
	kmeans = KMeans(n_clusters=k).fit(U)
	lab = kmeans.labels_


	pca = PCA(n_components=min(2,k)).fit(U)
	pca_2d = pca.transform(U)
	# print('pca_2d\n', pca_2d)

	if PRINT:

		f = plt.figure(0)
		plt.scatter(pca_2d[:, 0], pca_2d[:, 1])
		plt.title('Clustering visualized in 2D (using PCA)')
		plt.show()

		f = plt.figure(1)
		plt.plot(eigval)
		plt.title('Eigenvalues (sorted), with a sharp change at k={}'.format(k))
		plt.show()
		# kmeans.predict()
		# kmeans.cluster_centers_

		print('eigval', np.round(eigval, 2))
		print('lab', lab)

	return lab, k
